'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

exports.default = function (_ref) {
  var
  // Mising "types"
  //  - Comment
  //  - BlockComment
  //  - &attributes
  //  - InterpolatedTag
  //  - Code(buffer: false + block)
  //  - Each(object)
  //  - Mixin
  babel = _ref.babel;
  var parse = _ref.parse;
  var helpers = _ref.helpers;
  var ast = _ref.ast;
  var path = _ref.path;
  var t = babel.types;

  var compiler = {
    unwrap: function unwrap(node) {
      if (t.isJSXExpressionContainer(node)) {
        return node.expression;
      } else if (t.isJSXText(node)) {
        return t.stringLiteral(node.value);
      } else {
        return node;
      }
    },
    unwrapExpressionToStatements: function unwrapExpressionToStatements(node) {
      if (t.isDoExpression(node)) {
        return node.body.body;
      } else if (t.isJSXExpressionContainer(node)) {
        return this.unwrapExpressionToStatements(node.expression);
      } else {
        return [t.expressionStatement(this.unwrap(node))];
      }
    },
    parseExpression: function parseExpression(src) {
      var val = parse('x = (' + src + ');').program.body;
      (0, _assert2.default)(val.length === 1);
      return val[0].expression.right;
    },
    parseStatement: function parseStatement(src) {
      var val = parse(src).program.body;
      (0, _assert2.default)(val.length === 1);
      return val[0];
    },
    visit: function visit(node) {
      if (typeof this['visit' + node.type] === 'function') {
        return this['visit' + node.type](node);
      } else {
        throw new Error(node.type + ' is not yet supported');
      }
    },
    visitBlock: function visitBlock(node) {
      var _this = this;

      return node.nodes.map(function (node) {
        return _this.visit(node);
      }).filter(Boolean);
    },
    visitCase: function visitCase(node) {
      // compile the case/when into a series of conditionals then compile that down to jsx
      var base = { alternate: null };
      var parent = base;
      var defaultValue = { type: 'Block', nodes: [] };
      node.block.nodes.forEach(function (whenNode) {
        if (whenNode.expr === 'default') {
          defaultValue = whenNode.block;
        } else {
          parent.alternate = {
            type: 'Conditional',
            test: node.expr + ' === ' + whenNode.expr,
            consequent: whenNode.block,
            alternate: null
          };
          parent = parent.alternate;
        }
      });
      if (!base.alternate) {
        throw new Error('Empty case blocks are not supported');
      }
      parent.alternate = defaultValue;
      return this.visit(base.alternate);
    },
    visitCode: function visitCode(node) {
      if (node.buffer) {
        if (!node.mustEscape) {
          throw new Error('Unescaped, buffered code is not supported in react-pug');
        }
        return t.jSXExpressionContainer(this.parseExpression(node.val));
      } else {
        // TODO: hoist and rename `const` and `let` variables
        return t.jSXExpressionContainer(t.doExpression(t.blockStatement([this.parseStatement(node.val), t.expressionStatement(t.nullLiteral())])));
      }
    },
    simplifyBlock: function simplifyBlock(nodes) {
      var id = path.scope.generateUidIdentifier('SimpleBlock').name;
      nodes = nodes.map(this.unwrap);
      if (nodes.length === 1) {
        return nodes[0];
      }
      nodes.forEach(function (node, i) {
        if (t.isJSXElement(node) && !node.openingElement.attributes.some(function (attr) {
          return t.isJSXAttribute(attr) && t.isJSXIdentifier(attr.name, { name: 'key' });
        })) {
          var attr = t.jSXAttribute(t.jSXIdentifier('key'), t.stringLiteral(id + '_' + i));
          attr.pugSimpleBlockAttr = true;
          node.openingElement.attributes.push(attr);
        }
      });
      return t.arrayExpression(nodes);
    },
    visitConditional: function visitConditional(node) {
      var test = this.parseExpression(node.test);
      var consequent = this.simplifyBlock(this.visitBlock(node.consequent));
      var alternate = !node.alternate ? t.nullLiteral() : node.alternate.type === 'Block' ? this.simplifyBlock(this.visitBlock(node.alternate)) : this.unwrap(this.visitConditional(node.alternate));
      return t.jSXExpressionContainer(t.conditionalExpression(test, consequent, alternate));
    },
    simplifyDynamicBlock: function simplifyDynamicBlock(block /* Array<Expression> */, nodesIdentifier) {
      var _this2 = this;

      if (!nodesIdentifier) {
        throw new Error('Simplify Dynamic Block requires an identifier');
      }
      var body = [];
      var key = null;
      block.map(function (exp) {
        return _this2.unwrapExpressionToStatements(exp);
      }).forEach(function (statements, i) {
        var last = statements.pop();
        statements.forEach(function (s) {
          return body.push(s);
        });
        t.assertExpressionStatement(last);
        if (!t.isNullLiteral(last.expression)) {
          if (t.isJSXElement(last.expression)) {
            last.expression.openingElement.attributes = last.expression.openingElement.attributes.filter(function (attr) {
              return !attr.pugSimpleBlockAttr;
            });
            var hasKey = last.expression.openingElement.attributes.some(function (attr) {
              if (t.isJSXAttribute(attr) && t.isJSXIdentifier(attr.name, { name: 'key' })) {
                if (t.isJSXExpressionContainer(attr.value) && key === null) {
                  key = attr.value.expression;
                }
                return true;
              }
              return false;
            });
            if (!hasKey && key) {
              last.expression.openingElement.attributes.push(t.jSXAttribute(t.jSXIdentifier('key'), t.jSXExpressionContainer(t.binaryExpression('+', key, t.stringLiteral('_' + i)))));
            }
          }
          body.push(t.expressionStatement(t.callExpression(t.memberExpression(nodesIdentifier, t.identifier('push')), [last.expression])));
        }
      });
      return body;
    },
    visitEach: function visitEach(node) {
      var nodes = path.scope.generateUidIdentifier('EachBlock');
      var arr = this.parseExpression(node.obj);
      var params = [this.parseExpression(node.val)];
      if (node.key) params.push(this.parseExpression(node.key));
      var block = this.visitBlock(node.block);
      var body = block.length === 1 ? [t.returnStatement(this.unwrap(block[0]))] : [t.variableDeclaration('var', [t.variableDeclarator(nodes, t.arrayExpression([]))])].concat(this.simplifyDynamicBlock(block, nodes)).concat([t.returnStatement(nodes)]);

      var mapCall = t.callExpression(t.memberExpression(arr, t.identifier('map')), [t.functionExpression(null, params, t.blockStatement(body)), t.thisExpression()]);
      var condition = t.logicalExpression('&&', arr, t.memberExpression(arr, t.identifier('length')));
      var flatBody = body.length === 1 ? mapCall : t.callExpression(helpers.flatten, [mapCall]);
      var alternate = node.alternate ? this.simplifyBlock(this.visitBlock(node.alternate)) : t.nullLiteral();
      var loop = t.jSXExpressionContainer(t.conditionalExpression(condition, flatBody, alternate));
      return loop;
    },
    visitTag: function visitTag(node) {
      var _this3 = this;

      var name = t.jSXIdentifier(node.name);
      var children = (node.code ? [this.visitCode(node.code)] : []).concat(this.visitBlock(node.block)).reduce(function (a, b) {
        return a.concat(Array.isArray(b) ? b : [b]);
      }, []);

      if (node.attributeBlocks.length) {
        throw new Error('Attribute blocks are not yet supported in react-pug');
      }

      var classes = [];
      var attrs = node.attrs.map(function (_ref2) {
        var name = _ref2.name;
        var val = _ref2.val;
        var mustEscape = _ref2.mustEscape;

        if (/\.\.\./.test(name) && val === true) {
          return t.jSXSpreadAttribute(_this3.parseExpression(name.substr(3)));
        }
        switch (name) {
          case 'for':
            name = 'htmlFor';
            break;
          case 'maxlength':
            name = 'maxLength';
            break;
          case 'class':
            name = 'className';
            break;
        }
        var expr = _this3.parseExpression(val === true ? 'true' : val);
        if (!mustEscape && (!t.isStringLiteral(expr) || /(\<\>\&)/.test(val))) {
          throw new Error('Unescaped attributes are not supported in react-pug');
        }
        if (name === 'className') {
          classes.push(expr);
          return null;
        }
        var jsxValue = t.isStringLiteral(expr) || t.isJSXElement(expr) ? expr : t.jSXExpressionContainer(expr);
        if (/\.\.\./.test(name)) {
          throw new Error('spread attributes must not have a value');
        }
        return t.jSXAttribute(t.jSXIdentifier(name), jsxValue);
      }).filter(Boolean);
      if (classes.length) {
        var value = classes.every(function (cls) {
          return t.isStringLiteral(cls);
        }) ? t.stringLiteral(classes.map(function (cls) {
          return cls.value;
        }).join(' ')) : t.jSXExpressionContainer(t.callExpression(t.memberExpression(t.arrayExpression(classes), t.identifier('join')), [t.stringLiteral(' ')]));
        attrs.push(t.jSXAttribute(t.jSXIdentifier('className'), value));
      }
      var open = t.jSXOpeningElement(name, attrs, // Array<JSXAttribute | JSXSpreadAttribute>
      children.length === 0);
      var close = children.length === 0 ? null : t.jSXClosingElement(name);
      return t.jSXElement(open, close, children, // ["StringLiteral","JSXExpressionContainer","JSXElement"]
      children.length === 0);
    },
    visitText: function visitText(node) {
      return t.jSXText(node.val);
    },
    visitWhile: function visitWhile(node) {
      var nodes = path.scope.generateUidIdentifier('WhileBlock');
      var statements = [];
      statements.push(t.variableDeclaration('var', [t.variableDeclarator(nodes, t.arrayExpression([]))]));
      var test = this.parseExpression(node.test);
      var body = this.simplifyDynamicBlock(this.visitBlock(node.block), nodes);
      statements.push(t.whileStatement(test, t.blockStatement(body)));
      statements.push(t.expressionStatement(nodes));
      return t.jSXExpressionContainer(t.doExpression(t.blockStatement(statements)));
    }
  };
  return compiler.visit(ast);
};

var _assert = require('assert');

var _assert2 = _interopRequireDefault(_assert);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }